# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Anna Wojdel <awojdel@fluendo.com>

from elisa.core.media_uri import MediaUri
from elisa.core.common import application

from elisa.plugins.poblesec.actions import Action, \
                                           ContextualAction, \
                                           OpenControllerAction
                                           
from elisa.plugins.poblesec.actions import LinkAction

from elisa.plugins.database.actions import MoreOptionsAction, \
                                           EnqueueAndPlayMusicMixin

from elisa.plugins.grooveshark.utils import get_grooveshark_provider, \
                                            grooveshark_error_message, \
                                            stop_animation_on_error  

from elisa.plugins.grooveshark.models import GroovesharkFeedModel, \
                                             GroovesharkArtistModel, \
                                             GroovesharkAlbumModel, \
                                             GroovesharkPlaylistModel, \
                                             GroovesharkTrackModel

from elisa.core.utils import defer
from twisted.internet import task

from random import shuffle

from elisa.core.utils.i18n import install_translation

_ = install_translation('grooveshark')
_poblesec = install_translation('poblesec')


class SearchAction(LinkAction):
    """
    L{LinkAction} leading to L{GroovesharkSearcher}. 
    """
    path = '/poblesec/grooveshark/search/music'
    args = {'mediatype': 'music'}
    title = _('Search')


class AuthenticationAction(ContextualAction):
    
    def __init__(self, controller):
        super(AuthenticationAction, self).__init__(controller)
        provider = get_grooveshark_provider()
        self.logged = provider.authenticate_user()
        self.set_label(self.logged)

    def set_label(self, logged):
        if self.logged:
            self.name = _('Log out')
        else:
            self.name = _('Log in')

    def execute(self, item):
        def change_settings(result):
            self.logged = not self.logged
            self.set_label(self.logged)
        
        if self.logged:
            provider = get_grooveshark_provider()
            dfr = provider.logout()
            # Refresh settings only when performing logout
            # For login a new controler is appended, so refreshing is not needed
            dfr.addCallback(change_settings)
        else:
            browser = self.controller.frontend.retrieve_controllers('/poblesec/browser')[0]
            path = "/poblesec/grooveshark/login"
            dfr = browser.history.append_controller(path, self.name)
        return dfr



class LoginAction(LinkAction):
    """ 
    L{LinkAction} leading to the Grooveshark login form. Once the form has been 
    validated, it allows the user to go back to the Grooveshark plugin menu.
    """

    title = _('Login to Grooveshark')
    path = '/poblesec/grooveshark/login'
#    icon = 'elisa.plugins.poblesec.key_icon'


class LogoutAction(Action):
    """ 
    Action used to clear the Grooveshark login credentials. 
    It calls L{GroovesharkResourceProvider.logout} and returns 
    to the main Grooveshark menu.
    """

    title = _('Logout From Grooveshark')
#    icon = 'elisa.plugins.poblesec.remove'

    def run(self):
        def set_controllers(logout_response):
            browser = self.controller.frontend.retrieve_controllers('/poblesec/browser')[0]
            browser.history.go_back()
            browser.history.append_controller('/poblesec/grooveshark',
                                              'Grooveshark')
            return browser

        def logout_failed(failure):
            return defer.fail(failure)

        dfr = get_grooveshark_provider().logout()
        dfr.addCallback(set_controllers)
        dfr.addErrback(logout_failed)

        return dfr


class GroovesharkPlayAllActionMixin(ContextualAction, EnqueueAndPlayMusicMixin):
    
    """
    Action which plays all tracks which are of L{GroovesharkTrackModel} type.
    """
    def enqueue_and_play(self, items):
        # FIXME: this is not dynamic enough, one needs to first build the list
        # of PlayableModel's before actually starting to play the first item.
        # The playback of the first item should start first and then the list
        # of following items to enqueue in the playlist be built.
        if not items:
            return items
        
        self.controller.debug('start playing')
        
        def got_first_track(result):
            self.controller.debug('successfully downloaded first track')
            super(GroovesharkPlayAllActionMixin, self).enqueue_and_play(items)

        #dfr = items[0].prefetch_playable_model()
        dfr = defer.succeed(None)
        dfr.addCallbacks(got_first_track, grooveshark_error_message)
        return dfr

    def execute(self, item):
        if isinstance(item, GroovesharkTrackModel):
            tracks = self.controller.model
            index = tracks.index(item)
            dfr = defer.succeed(tracks[index:])
        else:
            dfr = item.get_tracks()
        dfr.addCallback(self.enqueue_and_play)
        return dfr


class GroovesharkShuffleAction(GroovesharkPlayAllActionMixin):    
    label = _('Shuffle And Play All Tracks')

    def execute(self, item):
        def got_tracks(tracks):
            shuffle(tracks)
            return tracks

        if isinstance(item, GroovesharkTrackModel):
            tracks = self.controller.tracks
            index = tracks.index(item)
            dfr = defer.succeed(tracks[index:])
        else:
            dfr = item.get_tracks()
        dfr.addCallback(got_tracks)
        dfr.addCallback(self.enqueue_and_play)
        return dfr


class GroovesharkTrackToFavoritesAction(ContextualAction):
    label = ''

    def __init__(self, controller):
        self.controller = controller
        self.favorite = self.controller.is_favorite
        if self.favorite:
            self.toggle(True)
        else:
            self.toggle(False)
        self.connecting = False
        return super(GroovesharkTrackToFavoritesAction, self).__init__(controller)

    def toggle(self, status):
        self.controller.debug("Toggling %s" % status)
        self.toggled = status
        self.refresh()

    def execute(self, item):
        if self.connecting:
            return defer.succeed(None)
        
        self.connecting = True

        if self.toggled:
            uri = 'grooveshark:///song/unfavorite'
        else:
            uri = 'grooveshark:///song/favorite'

        #FIXME:
        #See description in GroovesharkAddtrackToPlaylistAction
        #def refresh_icon(result, action)
        #    self.toggle(not self.toggled)
        #    self.controller.notifier.emit('favorites-status-changed', action) 
        def refresh_icon(result):
            self.toggle(not self.toggled)
            self.controller.debug('favourite status changed')
            self.controller.notifier.emit('status-changed')
            self.connecting = False

        return_model, dfr = application.resource_manager.post(MediaUri(uri), 
                                                        songID=item.song_id)
        dfr.addCallback(refresh_icon)
        dfr.addErrback(grooveshark_error_message)
        return dfr

    def refresh(self):
        if self.toggled:
            self.label = _('Remove From Favorites') 
        else:
            self.label = _('Add To Favorites')

        # visually refresh the list item
        # FIXME: uses reassignment to inform the widget that some values have
        # changed
        try:
            index = self.controller.model.index(self)
            self.controller.model[index] = self
        except ValueError, exception:
            # the action is not yet in the list of actions
            pass


class GroovesharkTrackToPlaylistsAction(ContextualAction):
    label = ''

    toggled = False
    playlist = None

    def __init__(self, controller, tracks=None):
        self.controller = controller
        self.tracks = tracks
        try:
            self.playlist = controller.playlist
        except AttributeError:
            self.playlist = None
        self.local_position = None
        self._toggle(self.playlist is not None)
    
        self.connecting = False
    
        return super(GroovesharkTrackToPlaylistsAction, self).__init__(controller)

    def _toggle(self, status):
        self.toggled = status
        self._refresh()

    def _refresh(self):
        if self.toggled:
            self.label = _('Remove From Playlist')
        else:
            self.label = _('Add To Playlist')
            
        # visually refresh the list item
        # FIXME: uses reassignment to inform the widget that some values have
        # changed
        try:
            index = self.controller.model.index(self)
            self.controller.model[index] = self
        except ValueError, exception:
            # the action is not yet in the list of actions
            pass

    def execute(self, item):
        if self.connecting:
            return defer.succeed(None)
        
        if self.playlist is not None:
            self.connecting = True
            """
            if "add/remove to playlist" action is triggered for song from a 
            given playlist, allow to add/remove this track only to/from this 
            playlist
            """

            #TODO:
            #In principle, there are two possible methods of dealing with 
            #changes introduced in the list of songs in playlist:
            #1) We can just notify any controller holding the list of songs 
            #that there was a change in the list, and let it contact 
            #Grooveshark to obtain an up-to-date playlist. In this case, all
            #we need to do is to emit a signal ('status-changed'), knowing that
            #interested controllers are listening to it.
            #2) We can avoid contacting Grooveshark server (which may take 
            #time) by updating the list of songs from playlist locally. In this
            #case, we need not only to emit a signal, but we also need to 
            #notify about the type of changes that happened.

            #At this moment, Grooveshark plugin uses the first method (always 
            #refreshing the list from the server), which results in a simple
            #'refresh_icon' function. In case we decide to move to solution (2),
            #the refresh_icon function should look more like the version 
            #commented out below. Note that this would also need appropriate 
            #changes in GroovesharkTracksController.node_clicked() where emitted 
            #signals are connected to appropriate controller and in 
            #GroovesharkTrackActionsController __gsignals__ definition.

            #def refresh_icon(result_model, action, position):
            #    self.toggle(not self.toggled)
            #    track = self.controller.track
            #    self.controller.notifier.emit('playlist-status-changed', 
            #                                   track, 
            #                                   action, position)

            #The whole discussion above is also relevant to updating list of 
            #favourite songs. In general, this can be viewed as a more general 
            #problem of "relying always on server to get the state of user's 
            #account" versus "building local knowledge of user's account". 
            #The first case is much simpler but involves more communication
            #with the server. It is what is implemented in the plugin at 
            #the moment.
            def refresh_icon(result_model):
                self._toggle(not self.toggled)
                self.controller.debug('playlist status changed')
                self.controller.notifier.emit('status-changed')
                if self.local_position is None:
                    self.local_position = len(self.tracks)
                self.connecting = False
                return defer.suceed(None)

            if self.toggled:
                uri = 'grooveshark:///playlist/removesong'
                if self.local_position is not None:
                    position = self.local_position
                else:
                    position = self.tracks.index(item) + 1

                post_arg = {'playlistID': self.playlist.playlist_id,
                            'position': unicode(position)}
                self.controller.debug('removing song from playlist at position %d',
                                      position)
            else:
                uri = 'grooveshark:///playlist/addsong'
                #FIXME:
                #The position parameter in playlist.addSong is not used. 
                #Currently, user can not set it from the plugin, so we assume 
                #that the song is always added at the end of the playlist. 
                #It is default 'behaviour' of the API, and therefore we don't 
                #need to fill position parameter with any value.
                post_arg = {'playlistID': self.playlist.playlist_id,
                            'songID': item.song_id,
                            'position': ''}
                self.controller.debug('adding song to playlist')

            return_model, dfr = application.resource_manager.post(MediaUri(uri),
                                                                  **post_arg)
            dfr.addCallbacks(refresh_icon, grooveshark_error_message)
        else:
            """
            if "add/remove to playlist" action is triggered for any song 
            (obtained from favorite, popular, or search) lists all available 
            playlists to choose a playlist to which the song should be added
            """
            playlist_browser = \
                self.controller.frontend.retrieve_controllers('/poblesec/browser')[0]
            path = '/poblesec/grooveshark/track/addtoplaylist'
            dfr = playlist_browser.history.append_controller(path, 
                                                             self.label, 
                                                             track=item)

        return dfr

class GroovesharkAddTrackToPlaylistAction(ContextualAction):
    label = _('Add To Playlist')
    
    def __init__(self, controller, track=None):
        self.controller = controller
        self.track = track
        return super(GroovesharkAddTrackToPlaylistAction, self).__init__(controller)
    
    def execute(self, item):
        def go_back(result_model):
            browser = self.controller.frontend.retrieve_controllers('/poblesec/browser')[0]
            return browser.history.go_back()

        uri = 'grooveshark:///playlist/addsong'
        #FIXME:
        #The position parameter in playlist.addSong is not used. 
        #Currently, user can not set it from the plugin, so we assume 
        #that the song is always added at the end of the playlist. 
        #It is default 'behaviour' of the API, and therefore we don't 
        #need to fill position parameter with any value.
        post_arg = {'playlistID': item.playlist_id,
                    'songID': self.track.song_id,
                    'position': ''}
        self.controller.debug('adding song with song_id=%s to playlist with playlist_id=%s', \
                                        self.track.song_id, item.playlist_id)
        return_model, dfr = application.resource_manager.post(MediaUri(uri), 
                                                              **post_arg)
        dfr.addCallbacks(go_back, grooveshark_error_message)
        return dfr


class GroovesharkBrowseAction(ContextualAction):
    """
    Default (all-in-one) action for the GroovesharkController, that allows
    to browse through grooveshark menu
    """
    def _run_action(self, item):
        return item.run()

    def _open_feed(self, item):
        browser = self.controller.frontend.retrieve_controllers('/poblesec/browser')[0]
        if item.label == 'songs':
            path = '/poblesec/grooveshark/tracks'
        elif item.label == 'artists':
            path = '/poblesec/grooveshark/artists'
        elif item.label == 'albums':
            path = '/poblesec/grooveshark/albums'
        elif item.label == 'playlists':
            path = '/poblesec/grooveshark/playlists'
        elif item.label == 'favorites':
            path = '/poblesec/grooveshark/tracks'
        else:
            path = self.controller.path

        dfr = browser.history.append_controller(path, item.title, uri=item.uri)
        dfr.addErrback(stop_animation_on_error, self.controller)
        return dfr

    def _open_tracks_controller(self, item, **args):
        browser = self.controller.frontend.retrieve_controllers('/poblesec/browser')[0]
        path = '/poblesec/grooveshark/tracks'
        dfr = browser.history.append_controller(path, item.name, **args)
        dfr.addErrback(stop_animation_on_error, self.controller)
        return dfr

    def execute(self, item):
        if isinstance(item, Action):
            return self._run_action(item)
        elif isinstance(item, GroovesharkFeedModel):
            return self._open_feed(item)
        elif isinstance(item, GroovesharkArtistModel):
            return self._open_tracks_controller(item, artist=item)
        elif isinstance(item, GroovesharkAlbumModel):
            return self._open_tracks_controller(item, album=item)
        elif isinstance(item, GroovesharkPlaylistModel):
            return self._open_tracks_controller(item, playlist=item)


class GroovesharkArtistMoreOptionsAction(MoreOptionsAction):
    path = '/poblesec/grooveshark/artist/more'

    def execute(self, item):
        return self.open_controller(self.path, item.name, artist=item)


class GroovesharkArtistTracksPlayAction(GroovesharkPlayAllActionMixin):
    label = _('Play All Tracks')


class GroovesharkAlbumTracksPlayAction(GroovesharkPlayAllActionMixin):
    pass


class GroovesharkListAlbumTracksAction(OpenControllerAction):
    name = 'list'

    def __init__(self, controller):
        path = '/poblesec/grooveshark/album/list'
        super(GroovesharkListAlbumTracksAction, self).__init__(controller, path)

    def execute(self, item):
        return self.open_controller(self.path, item.name, album=item)

        
class GroovesharkAlbumMoreOptionsAction(MoreOptionsAction):
    path = '/poblesec/grooveshark/album/more'

    def execute(self, item):
        return self.open_controller(self.path, item.name, album=item)

    
class GroovesharkTrackPlayAction(GroovesharkPlayAllActionMixin):
    pass


class GroovesharkTrackMoreOptionsAction(MoreOptionsAction):
    path = '/poblesec/grooveshark/track/more'

    def __init__(self, controller):
        self.parent_controller = controller
        try:
            self.parent_playlist = controller.playlist
        except AttributeError:
            self.parent_playlist = None
        self.parent_tracks = controller.model
        return super(GroovesharkTrackMoreOptionsAction, self).__init__(controller)

    def execute(self, item):
        def notify_parent_controller(controller):
            #FIXME: TODO not adapted for relook
            #TODO: option for local update of model
            #Together with the following functions it can be used to limit
            #the number of connections to grooveshark server. See details 
            #in description inside GroovesharkAddTrackToPlaylistAction 
            #in elisa.plugins.grooveshark.actions module 
            #
            #if self.contains_favorites:
            #    controller.list_controller.notifier.connect('favorites-status-changed', 
            #                                        self.refresh_favorites)
            #if self.contains_playlist:
            #   controller.list_controller.notifier.connect('playlist-status-changed',
            #                                        self.refresh_playlist)
            
            #FIXME: for unspecified reason in this callback controller argument 
            #is None, so we have to retrieve it independently
            controllers = self.parent_controller.frontend.retrieve_controllers('/poblesec/browser')
            browser = controllers[0]
            controller = browser.history.current
            
            controller.notifier.connect('status-changed', 
                                        self.parent_controller.refresh_model)
            
        dfr = self.open_controller(self.path, 
                                   item.title, 
                                   track=item, 
                                   playlist=self.parent_playlist, 
                                   tracks=self.parent_tracks)
        dfr.addCallback(notify_parent_controller)
        return dfr
