#-*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Anna Wojdel <awojdel@fluendo.com>

from elisa.core.common import application
from elisa.core.media_uri import MediaUri
from elisa.core.application import PICTURES_CACHE
from elisa.core.utils.i18n import install_translation
from elisa.core.utils import defer, caching

from elisa.plugins.base.utils import get_and_cache_image

from elisa.plugins.poblesec.base.list import GenericListViewMode, \
                                             BaseListController
from elisa.plugins.poblesec.base.grid import GridItemGridController
from elisa.plugins.poblesec.base.list_switcher import ListSwitcherController
from elisa.plugins.poblesec.base.preview_list import MenuItemPreviewListController
from elisa.plugins.poblesec.base.preview_list import DoubleLineMenuItemPreviewListController
from elisa.plugins.poblesec.base.info_list import InfoListController, \
                                                  ListInfoScreen
from elisa.plugins.poblesec.base.coverflow import ImageWithReflectionCoverflowController

from elisa.plugins.poblesec.link import Link

from elisa.plugins.poblesec.music_library import TracksViewMode
 
from elisa.plugins.database.music_controller import ArtistMoreOptionsController, \
                                                    AlbumInfoListController, \
                                                    AlbumInfoScreen

from elisa.plugins.database.tvshows_controllers import InfoListItemWidgetWithActions

from elisa.plugins.pigment.graph.text import Text
from elisa.plugins.pigment.widgets.list_vertical import ListVertical

from elisa.plugins.grooveshark.models import GroovesharkFeedModel
from elisa.plugins.grooveshark.models import GroovesharkArtistsListModel,  \
                                             GroovesharkArtistModel
from elisa.plugins.grooveshark.models import GroovesharkAlbumsListModel,  \
                                             GroovesharkAlbumModel
from elisa.plugins.grooveshark.models import GroovesharkTracksListModel,  \
                                             GroovesharkTrackModel
from elisa.plugins.grooveshark.models import GroovesharkPlaylistsListModel,  \
                                             GroovesharkPlaylistModel


from elisa.plugins.grooveshark.utils import get_grooveshark_provider
from elisa.plugins.grooveshark.utils import grooveshark_error_message

from elisa.plugins.poblesec.actions import Action
from elisa.plugins.grooveshark.actions import GroovesharkBrowseAction, \
                                              GroovesharkShuffleAction, \
                                              GroovesharkArtistMoreOptionsAction, \
                                              GroovesharkArtistTracksPlayAction, \
                                              GroovesharkAlbumTracksPlayAction, \
                                              GroovesharkListAlbumTracksAction, \
                                              GroovesharkAlbumMoreOptionsAction, \
                                              GroovesharkTrackPlayAction, \
                                              GroovesharkTrackMoreOptionsAction, \
                                              GroovesharkTrackToFavoritesAction, \
                                              GroovesharkTrackToPlaylistsAction, \
                                              GroovesharkAddTrackToPlaylistAction

from elisa.plugins.grooveshark.actions import SearchAction, \
                                              LoginAction, \
                                              LogoutAction

from elisa.plugins.grooveshark.player_controls import AutoPlayStartControl, \
                                                      AutoPlayHappyControl, \
                                                      AutoPlaySadControl


from elisa.plugins.pigment.widgets.theme import Theme
from elisa.plugins.pigment.graph.image import Image
from elisa.plugins.pigment.widgets.button import Button

import os
import gobject

from twisted.internet import reactor


_ = install_translation('grooveshark')
_poblesec = install_translation('poblesec')


def grooveshark_decorator(controller):
    controller.append_plugin( 'grooveshark',
                              _('Grooveshark'),
                              '/poblesec/grooveshark' )
    return defer.succeed(None)

#variables needed for proper managing of ribbon controls
start_autoplay_control = None
happy_control = None
sad_control = None

def grooveshark_player_decorator(controller):

    def set_ribbon_controls(player, model):
        """
        Method connected to 'current-model-changed' signal sent by player.
        Depending on the type of the model it adds or removes controls
        for autoplay feature supported only by Grooveshark web-service
        """
        global start_autoplay_control
        
        if isinstance(model, GroovesharkTrackModel):
            init_ribbon_controls(model)
            update_playlist()
        elif start_autoplay_control is not None: 
            remove_ribbon_controls()

    def init_ribbon_controls(model):
        """
        Called on entering the grooveshark mode for the player
        """                
        global start_autoplay_control
        global happy_control
        global sad_control

        ribbon = controller.player_osd.control_ribbon
        if start_autoplay_control is None:
            controller.debug('adding autoplay control to ribbon')
            start_autoplay_control = AutoPlayStartControl(controller)
            happy_control = AutoPlayHappyControl(controller)
            sad_control = AutoPlaySadControl(controller)

            ribbon.add_control(start_autoplay_control)
            start_autoplay_control.connect('glyphs-changed',update_happy_sad_controls)

        if model.autoplay_id is None:
            start_autoplay_control.set_state(False)

        return True

    def update_playlist():
        if start_autoplay_control.get_state():
            start_autoplay_control.update_autoplay_tracks()

    def remove_ribbon_controls():
        """
        Called on leaving the grooveshark mode of the player
        """
        global start_autoplay_control
        global happy_control
        global sad_control

        ribbon = controller.player_osd.control_ribbon

        controller.debug('removing autoplay control from ribbon')
        start_autoplay_control.set_state(False)
        ribbon.remove_control(start_autoplay_control)

        start_autoplay_control = None
        happy_control = None
        sad_control = None

        return True

    def update_happy_sad_controls(autoplay_control):
        ribbon = controller.player_osd.control_ribbon
        if autoplay_control.get_state():
            controller.debug('adding happy/sad controls to ribbon')
            ribbon.add_control(happy_control)
            ribbon.add_control(sad_control)
        else:
            controller.debug('removing happy/sad controls to ribbon')
            try:
                ribbon.remove_control(happy_control)
                ribbon.remove_control(sad_control)
            except:
                controller.debug('removing happy/sad controls failed')

    def playback_reached_30sec(player, stream_key, stream_server_id):
        """
        It sends information to grooveshark server that playback of song 
        reached 30 seconds.
        """
        #FIXME:
        #It does not check if playback really reached 30 seconds mark, but it is 
        #always called 30 seconds after playable_model was loaded. That means 
        #if there was delay (e.g. because of long buffering, or user clicked
        #pause, or even exited player) the 30 seconds mark will be reported 
        #to grooveshark (of course, only if the player still has the same 
        #playable_model, i.e. it is still potentially playing the song) 
        model = player.playlist.current_playable_model
        if hasattr(model, 'stream_key'):
            if model.stream_key == stream_key:
                uri = 'grooveshark:///song/reached30sec'
                result_model, dfr = \
                    application.resource_manager.post(uri=uri, 
                                                streamKey=stream_key,
                                                streamServerID=stream_server_id)
             
    def got_playable_model(player, playable_model):
        if hasattr(playable_model, 'stream_key') and \
                            hasattr(playable_model, 'stream_server_id'):
            reactor.callLater(30, playback_reached_30sec, 
                              player, 
                              stream_key=playable_model.stream_key, 
                              stream_server_id=playable_model.stream_server_id)
        
    controller.player.connect('current-model-changed', set_ribbon_controls)
    controller.player.connect('current-playable-model-changed', got_playable_model)
    return defer.succeed(None)


class GroovesharkNotifier(gobject.GObject):
    """
    This object emits three signals:
    favorites-status-changed - is emitted in case a song's favorite status is changed
    playlist-status-changed  - is emitted in case a song was added/removed to/from playlist
    status-changed           - is emitted together with both above signals
    Note that the first two signals are only used in case of local update of the 
    favorites/playlist lists, which is not currently in use.
    """
    __gsignals__ = {'favorites-status-changed':
                            (gobject.SIGNAL_RUN_LAST,
                            gobject.TYPE_BOOLEAN,
                            (gobject.TYPE_PYOBJECT, gobject.TYPE_STRING)),
                    'playlist-status-changed':
                            (gobject.SIGNAL_RUN_LAST,
                            gobject.TYPE_BOOLEAN,
                            (gobject.TYPE_PYOBJECT, gobject.TYPE_STRING, gobject.TYPE_INT)),
                    'status-changed':
                            (gobject.SIGNAL_RUN_LAST,
                            gobject.TYPE_BOOLEAN,
                            ()),
                   }


class GroovesharkController(BaseListController):
    """
    Grooveshark controller is responsible for those menu items in grooveshark 
    plugin, which do not require downloading data from grooveshark server.
    """

    start_uri = MediaUri('grooveshark://')
    popular_uri = MediaUri('grooveshark:///popular')

    def initialize(self, uri=start_uri):
        self.uri = uri
        self.debug('initializing GroovesharkController with uri = %s', str(uri))
        return super(GroovesharkController, self).initialize()
        
    def populate_model(self):
        provider = get_grooveshark_provider()
        user_logged = provider.authenticate_user()
        if user_logged is None:
            self.warning(_("User authentication hasn't finished yet"))

        model = []
        if self.uri == self.start_uri:
            model.append(SearchAction(self))
            feeds = [( _('Popular'), 
                       'popular', 
                       'popular')]
            if user_logged:
                feeds.extend( [ ( _('My Playlists'), 
                                  'user/playlists', 
                                  'playlists',), 
                                ( _('My Favorites'), 
                                  'user/favorites', 
                                  'favorites',) ] ) 

        elif self.uri == self.popular_uri:
            feeds = [( _('Artists'), 
                       'artists', 
                       'artists',), 
                     ( _('Albums'), 
                       'albums', 
                       'albums',),
                     ( _('Songs'), 
                       'songs', 
                       'songs',) ] 

        feeds_root = self.uri
        feed_models = []
        for title, feed_uri, label in feeds:
            feed_model = GroovesharkFeedModel()
            feed_model.title = title
            feed_model.uri = MediaUri(feeds_root + feed_uri)
            feed_model.label = label
            model.append(feed_model)

        if self.uri==self.start_uri:
            if user_logged:
                model.append(LogoutAction(self))
            else:
                model.append(LoginAction(self))

        return defer.succeed(model)

    def create_actions(self):
        return GroovesharkBrowseAction(self), []


class GroovesharkViewMode(GenericListViewMode):
    """
    Implementation of the common view for the Grooveshark Controller.
    """

    def get_label(self, item):
        return defer.succeed(item.title)
 
    def get_default_image(self, item):
        return None

    def get_preview_image(self, item, theme):
        return None
    
    def get_image(self, item, theme):
        return None
    

class GroovesharkPreviewListController(GroovesharkController, MenuItemPreviewListController):
    view_mode = GroovesharkViewMode
    item_widget_kwargs = {'with_artwork_box': False}
    crumbs_logo = "elisa.plugins.grooveshark.logo"


class GroovesharkList(ListVertical):
    def _render_widgets(self, start, end):
        if self._renderer is None:
            return

        for i, widget in enumerate(self._widgets[start:end+1]):
            item_index = self._item_index_from_widget_index(i+start)
            if item_index >= 0 and item_index < len(self.model):
                widget.visible = True
                item = self.model[item_index]
                # eventually cancel a previous render of the widget
                try:
                    previous_idle = self._render_calls.pop(item)
                except KeyError:
                    pass
                else:
                    gobject.source_remove(previous_idle)
                idle = gobject.idle_add(self._renderer, item, widget)
                self._render_calls[item] = idle
            else:
                widget.visible = False


class GroovesharkTracksControllerMixin(object):
    def create_actions(self):
        default = GroovesharkTrackPlayAction(self)

        provider = get_grooveshark_provider()
        if not provider.authenticate_user():
            return default, []

        more = GroovesharkTrackMoreOptionsAction(self)
        return default, [more]


class GroovesharkTracksController(GroovesharkTracksControllerMixin, 
                                  BaseListController):
    """
    Controller used for displaying lists of songs obtained from grooveshark server.
    """

    empty_label = _('There are no tracks in this section')

    songs_for_artist_uri = 'grooveshark:///artist/songs'
    songs_for_album_uri = 'grooveshark:///album/songs'
    songs_for_playlist_uri = 'grooveshark:///playlist/songs'


    def initialize(self, uri=None, artist=None, album=None, playlist=None, tracks=None, track=None):
        """
        Initialization of GroovesharkTracksController. Note that 'track' 
        argument to this function is not used - it is here only for 
        compatibility with Play All action from Database plugin.
        """
        self.artist = artist
        self.album = album
        self.playlist = playlist
        self.tracks = tracks
        self.track = track

        self.debug('initializing GroovesharkTracksController with uri=%s', str(uri))

        self.contains_favorites = False
        self.contains_playlist = False

        if uri is not None and uri.endswith('/favorites'):
            self.contains_favorites = True
        
        if artist is not None:
            self.uri = self.songs_for_artist_uri
            self.post_arg = {'artistID': artist.artist_id}
        elif album is not None:
            self.uri = self.songs_for_album_uri
            self.post_arg = {'albumID': album.album_id}
        elif playlist is not None:
            self.uri = self.songs_for_playlist_uri
            self.post_arg = {'playlistID': playlist.playlist_id}
            self.contains_playlist = True
        else:
            self.uri = uri
            self.post_arg = {}

        return super(GroovesharkTracksController, self).initialize()

    def populate_model(self):
        model = []

        def resource_loaded(resource):
            if isinstance(resource, GroovesharkTracksListModel):
                model.extend(resource.tracks)
            return model

        def load_resource():
            resource, get_deferred = \
                    application.resource_manager.post(self.uri, **self.post_arg)
            return get_deferred

        def extend_model(tracks):
            model.extend(tracks)
            return defer.succeed(model)
        
        if self.tracks is not None:
            dfr = extend_model(self.tracks)
        else:
            dfr = load_resource()
            dfr.addCallback(resource_loaded)
            dfr.addErrback(grooveshark_error_message)

        return dfr

    #OPTION for local update of model (currently not used):
    def refresh_favorites(self, controller, track, action):
        """
        Update model locally by adding/removing song directly to/from list
        """
        if action=='remove':
            self.model.remove(track)
        elif action=='add':
            self.model.insert(0,track)
        
        return self.model

    #OPTION for local update of model (currently not used):
    def refresh_playlist(self, controller, track, action, position):
        """
        Update model locally by adding/removing song directly to/from list
        """
        if action=='remove':
            self.model.remove(track)
        elif action=='add':
            if position > 0:
                index = position
            else: 
                index = len(self.model)
            self.model.insert(index, track)
        
        return self.model


    def refresh_model(self, controller):
        """
        Call provider to update model
        """
        def resource_loaded(resource):
            if isinstance(resource, GroovesharkTracksListModel):
                #reseting self.model is after call to resource_provider to 
				#minimise a time between setting model to empty and loading 
				#new tracks.
                self.model[:] = []
                self.model.extend(resource.tracks)
                self.widget.set_focus_proxy(self.nodes)
            return self.model

        if self.contains_playlist or self.contains_favorites:
            resource, dfr = application.resource_manager.post(uri=self.uri, **self.post_arg)
            dfr.addCallback(resource_loaded)
            dfr.addErrback(grooveshark_error_message)
        else:
            return defer.succeed(None)
        
        return dfr


class GroovesharkTracksViewMode(GenericListViewMode):
    """
    Implementation of the common view modes API.
    """
    def get_label(self, item):
        return defer.succeed(item.title)

    def get_sublabel(self, item):
        if isinstance(item, Action):
            return defer.succeed(item.subtitle)
        elif isinstance(item, GroovesharkTrackModel):
            subtitle = item.artist + ' (' + item.album.name + ')'
            return defer.succeed(subtitle)
        else:
            return None

    def get_default_image(self, item):
        return 'elisa.plugins.poblesec.glyphs.small.music'

    def get_image(self, item, theme):
        # Get a thumbnail for the track
        try:
            thumbnail_uri = item.thumbnails[0].references[0]
        except:
            return None

        # This will replace default thumbnail from Grooveshark 
        # website with one provided by Moovida
        if thumbnail_uri.endswith('default.png'):
            return None
        else:
            try:
                thumbnail_file = item.thumbnail_file
            except AttributeError:
                thumbnail_file = caching.get_cached_image_path(thumbnail_uri)
                item.thumbnail_file = thumbnail_file

            if os.path.exists(thumbnail_file):
                return defer.succeed(thumbnail_file)
            else:
                return get_and_cache_image(thumbnail_uri)

    def get_preview_image(self, item, theme):
        if isinstance(item, GroovesharkTrackModel):
            try:
                return item.thumbnail_file
            except AttributeError:
                return None
        else:
            return None


class GroovesharkTracksPreviewListController(GroovesharkTracksController, 
                                             DoubleLineMenuItemPreviewListController):
    list_widget = GroovesharkList
    view_mode = GroovesharkTracksViewMode
    crumbs_logo = "elisa.plugins.grooveshark.logo"


class TracksInfoScreen(ListInfoScreen):
    """
    Specialized ListInfoScreen widget to display information or more options for
    a music track.
    """

    def pack_captions(self, caption):
        caption.title = Text()
        caption.title.visible = True
        caption.foreground.pack_start(caption.title, expand=True)
        caption.album = Text()
        caption.album.visible = True
        caption.foreground.pack_start(caption.album, expand=True)
        caption.artist = Text()
        caption.artist.visible = True
        caption.foreground.pack_start(caption.artist, expand=True)


class GroovesharkTrackInfoScreen(InfoListController):
    view_mode = TracksViewMode
    info_screen_cls = TracksInfoScreen

    def initialize(self, track=None, playlist=None):
        self.track = track
        return super(GroovesharkTrackInfoScreen, self).initialize()

    def populate_info(self):
        default_resource = 'elisa.plugins.poblesec.glyphs.large.music'
        artwork = self.info_screen.left_panel.artwork

        self.info_screen.left_panel.caption.title.label = self.track.title
        self.info_screen.left_panel.caption.album.label = self.track.album.name
        self.info_screen.left_panel.caption.artist.label = self.track.artist

        def _set_poster(thumbnail_path):
            return artwork.foreground.set_from_file_deferred(thumbnail_path)

        try:
            thumbnail_uri = self.track.thumbnails[0].references[0]
        except:
            return self.frontend.load_from_theme(default_resource, 
                                                 artwork.foreground)

        # This will replace default thumbnail from Grooveshark 
        # website with one provided by Moovida
        if thumbnail_uri.endswith('default.png'):
            return self.frontend.load_from_theme(default_resource, 
                                                 artwork.foreground)
        else:
            dfr = get_and_cache_image(thumbnail_uri)
            dfr.addCallback(_set_poster)
            return dfr


class GroovesharkTrackMoreOptionsController(GroovesharkTrackInfoScreen):

    def initialize(self, track, playlist=None, tracks=None):
        self.playlist = playlist
        self.tracks = tracks
        
        self.is_favorite = False
        self.notifier = GroovesharkNotifier()

        def check_if_favorite(result_model):
            for item in result_model.tracks:
                if item.song_id==track.song_id:
                    self.is_favorite = True
                    break
            return self.is_favorite

        def load_favorite_songs():
            uri = 'grooveshark:///user/favorites'
            result_model, dfr = application.resource_manager.post(MediaUri(uri))
            return dfr

        def call_super(arg):
            return super(GroovesharkTrackMoreOptionsController, self).initialize(track)

        if get_grooveshark_provider().authenticate_user():
            self.debug('checking if song is favorited')
            dfr = load_favorite_songs()
            dfr.addCallback(check_if_favorite)
        else:
            dfr = defer.succeed(None)

        dfr.addCallback(call_super)
        
        return dfr

    def populate_info(self):
        self.info_screen.right_panel.title.foreground.label = _('MORE OPTIONS')
        return super(GroovesharkTrackMoreOptionsController, self).populate_info()
    
    def populate_model(self):
        add_to_favorites = GroovesharkTrackToFavoritesAction(self)
        add_to_playlist = GroovesharkTrackToPlaylistsAction(self, self.tracks)
        model = [add_to_favorites, add_to_playlist]
        return defer.succeed(model)

    def node_renderer(self, item, widget):
        widget.item_widget.label.label = item.label

    def item_activated(self, item):
        return item.execute(self.track)

    
class GroovesharkAddTrackToPlaylistController(GroovesharkTrackInfoScreen):
    empty_label = _('Please create a playlist at Grooveshark.com') 
         
    def populate_info(self):
        self.info_screen.right_panel.title.foreground.label = _('PLAYLISTS OPTIONS')
        return super(GroovesharkAddTrackToPlaylistController, self).populate_info()
    
    def populate_model(self):
        model = []
        
        def resource_loaded(resource):
            model.extend(resource.playlists)
            return defer.succeed(model)

        def load_resource():
            resource, get_deferred = application.resource_manager.post('grooveshark:///user/playlists')
            return get_deferred
        
        dfr = load_resource()
        dfr.addCallback(resource_loaded)
        dfr.addErrback(grooveshark_error_message)
        return dfr

    def node_renderer(self, item, widget):
        widget.item_widget.label.label = item.name

    def create_actions(self):
        return GroovesharkAddTrackToPlaylistAction(self, self.track), []

    #We have to override display_empty_alert to hide info_screen widget 
    def display_empty_alert(self, label):
        self.info_screen.visible = False
        super(GroovesharkAddTrackToPlaylistController, self).display_empty_alert(label)

class GroovesharkArtistsController(BaseListController):
    """
    Controller displaying a list of artists retrieved from grooveshark server 
    (either from search, or from popular selection).
    """

    empty_label = _('There are no artists in this section')

    def initialize(self, uri=None, artists=None):
        self.uri = uri
        self.artists = artists
        self.debug('initializing GroovesharkArtistController with uri=%s', str(uri))
        return super(GroovesharkArtistsController, self).initialize()

    def populate_model(self):
        model = []
        
        def resource_loaded(resource):
            if isinstance(resource, GroovesharkArtistsListModel):
                model.extend(resource.artists)
            return model

        def load_resource():
            resource, get_deferred = application.resource_manager.post(self.uri)
            return get_deferred
        
        def extend_model(artists):
            model.extend(artists)
            return defer.succeed(model)
        
        if self.artists is not None:
            return extend_model(self.artists)
        else:
            deferred = load_resource()
            deferred.addCallback(resource_loaded)
            deferred.addErrback(grooveshark_error_message)
            return deferred

    def create_actions(self):
        default = GroovesharkBrowseAction(self)
        # Contextual actions: more options.
        more = GroovesharkArtistMoreOptionsAction(self)
        return default, [more]


class GroovesharkArtistsViewMode(GenericListViewMode):
    """
    Implementation of view for grooveshark artist model.
    """

    def get_label(self, item):
        return defer.succeed(item.name)

    def get_default_image(self, item):
        return 'elisa.plugins.poblesec.glyphs.small.artist'

    def get_image(self, item, theme):
        return None
    
    def get_preview_image(self, item, theme):
        return None


class GroovesharkArtistsPreviewListController(GroovesharkArtistsController, MenuItemPreviewListController):
    view_mode = GroovesharkArtistsViewMode
    crumbs_logo = "elisa.plugins.grooveshark.logo"


class GroovesharkArtistMoreOptionsController(ArtistMoreOptionsController):
    
    def populate_info(self):
        default_resource = 'elisa.plugins.poblesec.glyphs.large.artist'
        artwork = self.info_screen.left_panel.artwork

        self.info_screen.right_panel.title.foreground.label = _('MORE OPTIONS')
        self.info_screen.left_panel.caption.artist.label = self.artist.name

        def _set_poster(thumbnail_path):
            return artwork.foreground.set_from_file_deferred(thumbnail_path)

        try:
            thumbnail_uri = self.track.thumbnails[0].references[0]
        except:
            return self.frontend.load_from_theme(default_resource, 
                                                 artwork.foreground)

        # This will replace default thumbnail from Grooveshark 
        # website with one provided by Moovida
        if thumbnail_uri.endswith('default.png'):
            return self.frontend.load_from_theme(default_resource, 
                                                 artwork.foreground)
        else:
            dfr = get_and_cache_image(thumbnail_uri)
            dfr.addCallback(_set_poster)
            return dfr

    def populate_model(self):
        play_all = GroovesharkArtistTracksPlayAction(self)
        shuffle = GroovesharkShuffleAction(self)
        model = [play_all, shuffle]
        return defer.succeed(model)
        

class GroovesharkAlbumsController(BaseListController):
    """
    Controller displaying a list of albums retrieved from grooveshark server 
    (either from search, or from popular selection).
    """

    empty_label = _('There are no albums in this section')

    def initialize(self, uri=None, albums=None):
        self.uri = uri
        self.albums = albums
        self.debug('initializing GroovesharkAlbumController with uri=%s', str(uri))
        return super(GroovesharkAlbumsController, self).initialize()

    def populate_model(self):
        model = []
        
        def resource_loaded(resource):
            if isinstance(resource, GroovesharkAlbumsListModel):
                model.extend(resource.albums)
            return model

        def load_resource():
            resource, get_deferred = application.resource_manager.post(self.uri)
            return get_deferred

        def extend_model(albums):
            model.extend(albums)
            return defer.succeed(model) 

        if self.albums is not None:
            dfr = extend_model(self.albums)
        else:
            dfr = load_resource()
            dfr.addCallback(resource_loaded)
            dfr.addErrback(grooveshark_error_message)

        return dfr

    def create_actions(self):
        default = GroovesharkAlbumTracksPlayAction(self)
        # Contextual actions: list tracks, more options.
        list_tracks = GroovesharkListAlbumTracksAction(self)
        more = GroovesharkAlbumMoreOptionsAction(self)
        return default, [list_tracks, more]


class GroovesharkAlbumsViewMode(GenericListViewMode):
    """
    Implementation of view for grooveshark artist model.
    """

    def get_label(self, item):
        return defer.succeed(item.name)

    def get_sublabel(self, item):
         if isinstance(item, GroovesharkAlbumModel):
            return defer.succeed(item.artist)
         else:
            return None
        
    def get_default_image(self, item):
        return 'elisa.plugins.poblesec.glyphs.small.music'

    def get_image(self, item, theme):
        # Get an album's cover
        try:
            thumbnail_uri = item.cover.references[0]
        except:
            return None

        # This will replace default thumbnail from Grooveshark
        # website with one provided by Moovida
        if thumbnail_uri.endswith('default.png'):
            return None
        else:
            try:
                thumbnail_file = item.thumbnail_file
            except AttributeError:
                thumbnail_file = caching.get_cached_image_path(thumbnail_uri)
                item.thumbnail_file = thumbnail_file

            if os.path.exists(thumbnail_file):
                return defer.succeed(thumbnail_file)
            else:
                return get_and_cache_image(thumbnail_uri)

    def get_preview_image(self, item, theme):
        if isinstance(item, GroovesharkAlbumModel):
            try:
                return item.thumbnail_file
            except AttributeError:
                return None
        else:
            return None


class GroovesharkAlbumsPreviewListController(GroovesharkAlbumsController, DoubleLineMenuItemPreviewListController):
    view_mode = GroovesharkAlbumsViewMode
    crumbs_logo = "elisa.plugins.grooveshark.logo"


class GroovesharkAlbumInfoScreenWithContextualActions(AlbumInfoScreen):
    node_widget = InfoListItemWidgetWithActions


class GroovesharkAlbumInfoListController(AlbumInfoListController):
    def populate_info(self):
        default_resource = 'elisa.plugins.poblesec.glyphs.large.music'
        artwork = self.info_screen.left_panel.artwork
        
        self.info_screen.left_panel.caption.album.label = self.album.name

        def _set_poster(thumbnail_path):
            return artwork.foreground.set_from_file_deferred(thumbnail_path)

        def got_artist_name(artist_name):
            self.info_screen.left_panel.caption.artist.label = artist_name
        
            # Get an album's cover
            try:
                thumbnail_uri = self.album.cover.references[0]
            except:
                return self.frontend.load_from_theme(default_resource, 
                                                     artwork.foreground)

            # This will replace default thumbnail from Grooveshark 
            # website with one provided by Moovida
            if thumbnail_uri.endswith('default.png'):
                return self.frontend.load_from_theme(default_resource, 
                                                     artwork.foreground)
            else:
                dfr = get_and_cache_image(thumbnail_uri)
                dfr.addCallback(_set_poster)
                return dfr

        dfr = self.album.get_artist_name()
        dfr.addCallback(got_artist_name)
        return dfr


class GroovesharkAlbumTracksListController(GroovesharkTracksControllerMixin, 
                                           GroovesharkAlbumInfoListController):
    
    info_screen_cls = GroovesharkAlbumInfoScreenWithContextualActions
    
    def populate_info(self):
        self.info_screen.right_panel.title.foreground.label = _('TRACKS')
        super(GroovesharkAlbumTracksListController, self).populate_info()
        
    def populate_model(self):
        return self.album.get_tracks()

    def node_renderer(self, item, widget):
        trash_letters = " .,-_"
        title = item.title.strip(trash_letters)
        if item.track_number:
            title = '%s. %s' % (str(item.track_number).zfill(2), title)
        widget.item = item
        widget.item_widget.label.label = title

    def refresh_model(self, controller):
        #we don't need to refresh model for popular albums
        pass

class GroovesharkAlbumMoreOptionsController(GroovesharkAlbumInfoListController):
    
    def populate_info(self):
        self.info_screen.right_panel.title.foreground.label = _('MORE OPTIONS')
        super(GroovesharkAlbumMoreOptionsController, self).populate_info()

    def populate_model(self):
        shuffle = GroovesharkShuffleAction(self)
        model = [shuffle]
        return defer.succeed(model)

    def node_renderer(self, item, widget):
        widget.item_widget.label.label = item.label

    def item_activated(self, item):
        return item.execute(self.album)


class GroovesharkPlaylistsController(BaseListController):
    """
    Controller displaying a list of playlists retrieved from grooveshark server 
    (either from search, or from popular selection).
    """

    empty_label = _('There are no playlists in this section')

    default_uri = 'grooveshark:///playlist/songs'

    def initialize(self, uri=default_uri):
        self.uri = uri
        self.debug('initializing GroovesharkPlaylistController with uri=%s', str(uri))
        return super(GroovesharkPlaylistsController, self).initialize()

    def populate_model(self):
        model = []
        
        def resource_loaded(resource):
            if isinstance(resource, GroovesharkPlaylistsListModel):
                model.extend(resource.playlists)
            return defer.succeed(model)

        def load_resource():
            resource, get_deferred = application.resource_manager.post(self.uri)
            return get_deferred
        
        dfr = load_resource()
        dfr.addCallback(resource_loaded)
        dfr.addErrback(grooveshark_error_message)

        return dfr

    def create_actions(self):
        return GroovesharkBrowseAction(self), []


class GroovesharkPlaylistsViewMode(GenericListViewMode):
    """
    Implementation of view for grooveshark artist model.
    """

    def get_label(self, item):
        return defer.succeed(item.name)

    def get_default_image(self, item):
        return None

    def get_image(self, item, theme):
        return None

    def get_preview_image(self, item, theme):
        return None


class GroovesharkPlaylistsPreviewListController(GroovesharkPlaylistsController, MenuItemPreviewListController):
    view_mode = GroovesharkPlaylistsViewMode
    item_widget_kwargs = {'with_artwork_box': False}
    crumbs_logo = "elisa.plugins.grooveshark.logo"


def use_me_hook(frontend):
    """
    'Use me' hook that takes the user to the Internet Media/Music/Grooveshark 
    section.

    @param frontend: the current frontend (poblesec)
    @type frontend:  L{elisa.plugins.pigment.pigment_frontend.PigmentFrontend}

    @return:         a deferred fired when the action is complete
    @rtype:          L{twisted.internet.defer.Deferred}
    """

    browser = frontend.retrieve_controllers('/poblesec/browser')[0]

    paths = [('/poblesec/internet_menu', _poblesec('INTERNET MEDIA'), {}),
             ('/poblesec/music/internet', _poblesec('Music'), {}),
             ('/poblesec/grooveshark', _('Grooveshark'), {})]

    return browser.navigate(paths)
