# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Anna Wojdel <awojdel@fluendo.com>

from elisa.core import common

from elisa.core.components.model import Model

from elisa.plugins.base.models.audio import ArtistModel, \
                                            AlbumModel, \
                                            TrackModel

from elisa.core.media_uri import MediaUri

from elisa.core.utils import defer

from elisa.plugins.http_client.http_client import ElisaAdvancedHttpClient
from twisted.web2 import responsecode

from elisa.core.utils.i18n import install_translation

_ = install_translation('grooveshark')
_poblesec = install_translation('poblesec')


class DummyModel(object):
    pass 


class GroovesharkFeedModel(Model):
    """
    This model contains feed data.

    @ivar title:          the title of the feed
    @type title:          C{unicode}
    @ivar uri:            the uri of the feed
    @type uri:            C{unicode}
    @ivar label:          label of the feed which is used for identification
    @type label:          C{unicode}
    @ivar resource:       image resource to be displayed for the item
    @type resource:       C{unicode}
    """

    def __init__(self):
        """
        Constructor. Initialize all the fields.
        """
        super(GroovesharkFeedModel, self).__init__()
        self.title = None
        self.uri = None
        self.label = None
        self.resource = None    #not used any more

class GroovesharkArtistsListModel(Model):
    """
    Artist list model contains a list of artist models.

    @ivar artists: the list of artists
    @type artists: C{list} of L{elisa.plugins.grooveshark.models.GroovesharkArtistModel}
    """

    def __init__(self):
        """
        Constructor. Initialize all the fields.
        """
        super(GroovesharkArtistsListModel, self).__init__()
        self.artists = []


class GroovesharkArtistModel(ArtistModel):
    """
    Grooveshark representation of an artist.

    @ivar artist_id:        grooveshark id of the artist
    @type artist_id:        C{int}
    """

    def __init__(self):
        super(GroovesharkArtistModel, self).__init__()

        self.artist_id = None

    def get_tracks(self):
        def resource_loaded(resource):
            return resource.tracks

        uri = 'grooveshark:///artist/songs'
        post_arg = {'artistID': self.artist_id}
        recourse, dfr = common.application.resource_manager.post(uri, **post_arg)
        dfr.addCallback(resource_loaded)
        return dfr

class GroovesharkAlbumsListModel(Model):
    """
    Album list model contains a list of album models.

    @ivar albums:    the list of albums
    @type albums:    C{list} of L{elisa.plugins.grooveshark.models.GroovesharkAlbumModel}
    """

    def __init__(self):
        """
        Constructor. Initialize all the fields.
        """
        super(GroovesharkAlbumsListModel, self).__init__()
        self.albums = []


class GroovesharkAlbumModel(AlbumModel):
    """
    Grooveshark representation of an album.

    @ivar album_id:       grooveshark id of the album
    @type album_id:       C{int}
    @ivar artist_id:      grooveshark id of the artist (not really used right now)
    @type artist_id:      C{int}
    """

    def __init__(self):
        super(GroovesharkAlbumModel, self).__init__()
        
        self.album_id = None
        self.artist_id = None
        
    def get_tracks(self):
        def resource_loaded(resource):
            return resource.tracks

        uri = 'grooveshark:///album/songs'
        post_arg = {'albumID': self.album_id}
        recourse, dfr = common.application.resource_manager.post(uri, **post_arg)
        dfr.addCallback(resource_loaded)
        return dfr


class GroovesharkTracksListModel(Model):
    """
    A track list model contains a list of track models.

    @ivar tracks: the list of tracks
    @type tracks: C{list} of L{elisa.plugins.grooveshark.models.GroovesharkTrackModel}
    """

    def __init__(self):
        """
        Constructor. Initialize all the fields.
        """
        super(GroovesharkTracksListModel, self).__init__()
        self.tracks = []


class GroovesharkTrackModel(TrackModel):
    """
    Grooveshark representation of the track.

    @ivar thumbnails:   list of thumbnails for song
    @type thumbnails:   C{list} of L{elisa.plugins.base.models.image.ImageModel}

    @ivar song_id:      the ID of the song in grooveshark database - needed for getting 
                        stream uri which is generated temporally 
    @type song_id:      C{unicode}
	@ivar autoplay_id:  the ID of the Autoplay session that the track belongs to
    @type autoplay_id:  C{int}
    
    @ivar _prefetched_model: L{elisa.plugins.base.models.media.PlayableModel} prefetched 
                             by L{self.prefetch_playable_model()} method.
    @type _prefetched_model: C{elisa.plugins.base.models.media.PlayableModel}
    """

    def __init__(self):
        super(GroovesharkTrackModel, self).__init__()
        self.thumbnails = []
        self.song_id = None
        self.autoplay_id = None
        self._prefetched_model = None

    def _prepare_playable_model(self, model):
        model.title = self.title
        model.allow_seek = False
        return model

    def _retrieve_playable_model(self):
        uri = 'grooveshark:///song/stream'
        return_model, dfr = common.application.resource_manager.post(uri, songID = self.song_id )
        dfr.addCallback(self._prepare_playable_model)
        return dfr

    def prefetch_playable_model(self):
        """
        Return a deferred that will return an instance of
        L{elisa.plugins.base.models.media.PlayableModel} for the Grooveshark
        song.
        This function is called by get_playable_model(), but it can be also
        called in advance in order to smooth the user experience. Prefetched 
        model is stored in self._prefetched_model.
        Note that the returned model should be passed to player within a short 
        time, before it becomes invalid.
        
        @rtype:  L{elisa.core.utils.defer.Deferred}
        """

        #FIXME:
        #There is a problem with Grooveshark collection, many of the songs
        #return streamableUrl which points to nowhere (HTTP 404). Unfortunately,
        #neither Player, nor gstreamer check for the response code from the 
        #HTTP request, and so the user is presented with "cannot play this 
        #stream" type of message.
        #We can check for the stream availability here, by making simple HTTP 
        #request, but as Grooveshark streamable URLs are one time only, 
        #there is no way to reuse the URL in the player.
        #Below is ugly workaround, which upon successful retrieval of stream, 
        #calls getStreamableUrl again and passes it to the player.
        #The proper solution should be to implement handling of HTTP response 
        #codes somewhere in Gstreamer - in such a case the code would look like
        #he one shown in comments below this function. 
        def cache_playable_model(model):
            self.debug('caching playable model')
            self._prefetched_model = self._prepare_playable_model(model)
            return model
        
        def stream_request_done(response):
            if response.code == responsecode.OK:
                dfr = self._retrieve_playable_model()
                dfr.addCallback(cache_playable_model)
            else:
                error_text = _('Server returned HTTP code: %d\n') % response.code
                if response.code == 404:
                    error_text += _('File not found on server')

                dfr = defer.fail(Exception('%s' % error_text))
            return dfr

        def received_playable_model(model):
            http_client = ElisaAdvancedHttpClient(host=model.uri.host)
            dfr = http_client.request(model.uri)
            dfr.addCallback(stream_request_done)
            return dfr

        dfr = self._retrieve_playable_model()
        dfr.addCallback(received_playable_model)
        
        return dfr
    """
    This is how this function should really look like:
    
    def prefetch_playable_model(self):
        def parse_model(model):
            model.title = self.title
            model.allow_seek = False
            
            self._prefetched_model = model
            return model
                    
        uri = 'grooveshark:///song/stream'
        return_model, dfr = common.application.resource_manager.post(uri, songID = self.song_id )
        dfr.addCallback(parse_model)
        return dfr
    """

    def get_playable_model(self):
        """
        Return a deferred that will return an instance of
        L{elisa.plugins.base.models.media.PlayableModel} for the Grooveshark song.
        
        @rtype:  L{elisa.core.utils.defer.Deferred}
        """
        return self._retrieve_playable_model()
        def consume_prefetched_model(model):
            self.debug('deleting cached model')
            self._prefetched_model = None
            return model

        if self._prefetched_model is not None:
            self.debug('returning prefetched playable model')
            dfr = defer.succeed(self._prefetched_model)
        else:
            self.debug('prefetching playable model')
            dfr = self.prefetch_playable_model()
            
        dfr.addCallback(consume_prefetched_model)
        return dfr


class GroovesharkPlaylistsListModel(Model):
    """
    A playlist list model contains a list of playlist models.

    @ivar playlists:    the list of playlists
    @type playlists:    C{list} of L{elisa.plugins.grooveshark.models.GroovesharkPlaylistModel}
    """

    def __init__(self):
        """
        Constructor. Initialize all fields.
        """
        super(GroovesharkPlaylistsListModel, self).__init__()
        self.playlists = []


class GroovesharkPlaylistModel(Model):
    """
    Representation of an playlist on grooveshark webservice.

    @ivar name:         name of the playlist
    @type name:         C{unicode}
    @ivar playlist_id:  the ID of the playlist in grooveshark database
    @type playlist_id:  C{int}
    @ivar num_songs:    number of songs in playlist 
    @type num_songs:    C{int}
    """

    def __init__(self):
        """
        Constructor. Initialize all the fields.
        """
        super(GroovesharkPlaylistModel, self).__init__()
        self.name = None
        self.playlist_id = None
        self.num_songs = None
        

class GroovesharkLoginModel(Model):
    """
    Login model containing user specific login information
    
    @ivar success:      state of the login (equal True if user is logged in)  
    @type success:      C{boolean}
    
    @ivar username:     username of the user (used for login to grooveshark webservice)
    @type username:     C{unicode}
    @ivar password:     password of the user (used for login to grooveshark webservice)
    @type password:     C{unicode}
    
    @ivar user_id:      user ID (returned by grooveshark API)
    @type user_id:      C{int}
    """

    def __init__(self):
        super(GroovesharkLoginModel, self).__init__()
        self.success = False
        self.username = None
        self.password = None
        self.user_id = None

    def reset(self):
        """
        Resets all variables to initial values
        """
        self.success = False
        self.username = None
        self.password = None
        self.user_id = None

