# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Anna Wojdel <awojdel@fluendo.com>

from elisa.plugins.poblesec.widgets.player.control_ribbon import Control
from elisa.plugins.poblesec.videoplayer_controls import MultiStateControl
from elisa.plugins.pigment.widgets.const import *

from elisa.plugins.grooveshark.models import GroovesharkTrackModel
from elisa.plugins.grooveshark.utils import grooveshark_error_message

from elisa.core.common import application
from elisa.core.utils import defer

from elisa.core.utils.i18n import install_translation


_ = install_translation('grooveshark')

class AutoPlayStartControl(MultiStateControl):
    """
    Start/Stop the autoplay feature.

    Its L{slave} passed to the constructor must be of type
    L{elisa.plugins.poblesec.player_video.PlayerController}.
    """
    def __init__(self, *args):
        super(AutoPlayStartControl, self).__init__(*args)
        # We call the superclass function to make sure that all
        # glyphs, captions and self._state are properly set, but without
        # all of the logic related to starting/stopping the autoplay 
        # session.
        self.disabled = False
        super(AutoPlayStartControl, self).set_state(False)
        
    _glyphs_table = {True: \
            {STATE_NORMAL: 'elisa.plugins.grooveshark.autoplay_check_normal',
             STATE_SELECTED: 'elisa.plugins.grooveshark.autoplay_check_selected',
             STATE_PRESSED: 'elisa.plugins.grooveshark.autoplay_check_active',
            },
            False: \
            {STATE_NORMAL: 'elisa.plugins.grooveshark.autoplay_normal',
             STATE_SELECTED: 'elisa.plugins.grooveshark.autoplay_selected',
             STATE_PRESSED: 'elisa.plugins.grooveshark.autoplay_active',
            }}

    _caption_table = {False: _('AutoPlay On'),
                     True: _('AutoPlay Off')}

    def set_state(self, status):
        """
        Start/stop autoplay session. Depending on the 'status' argument
        this function calls either L{self.start_autoplay_session} or
        L{self.stop_autoplay_session} and updates the glyphs for the control
        upon successful completion of the returned defers.

        @param status: enable or disable the autoplay feature
        @type status:  boolean
        """
        def refresh(autoplay):
            super(AutoPlayStartControl, self).set_state(status)
            self.disabled = False
        
        if status == self.get_state():
            return

        if status:
            dfr = self._start_autoplay_session()
        else:
            dfr = self._stop_autoplay_session()

        self.disabled = True
        dfr.addCallback(refresh)
        return dfr

    def _autoplay_fail(self, failure):
        grooveshark_error_message(failure)
        # This is an emergency action. We have to stop autoplay session
        # and set glyphs to inactive independently from actual state of 
        # the control. 
        # We need to call _stop_autoplay_session here because Grooveshark 
        # servers often return "internal error" on starting autoplay 
        # session - but after starting it (i.e. Grooveshark thinks we have
        # autoplay session, but we were not notified at all). 
        self._stop_autoplay_session()
        super(AutoPlayStartControl, self).set_state(False)
        return defer.fail(Exception('AutoPlay failed'))

    def update_autoplay_tracks(self):
        def update_playlist(result_model):
            if isinstance( result_model, GroovesharkTrackModel ):
                self.slave.player.enqueue_to_playlist(result_model)

        if self.slave.player.playlist.current_index == len(self.slave.player.playlist)-1:
            uri = 'grooveshark:///autoplay/getnextsong'
            result_model, dfr = application.resource_manager.post(uri)
            dfr.addCallbacks(update_playlist, self._autoplay_fail)


    def _start_autoplay_session(self):
        def update_model(result_model):
            if isinstance(result_model, GroovesharkTrackModel):
                playlist = self.slave.player.playlist
                current_index = playlist.current_index
                self.slave.player.playlist[current_index+1:] = []

                self.slave.player.enqueue_to_playlist( result_model )

        uri = 'grooveshark:///autoplay/start'
        current_track = self.slave.player.get_current_model()
        song_ids = current_track.song_id
        #Should we pass the whole playlist as context?
        #current_index = self.slave.player.playlist.current_index
        #song_ids = [int(item.song_id) for item in 
        #                       self.slave.player.playlist[:current_index+1] ]
        #or 
        #song_ids = [int(item.song_id) for item in 
        #                       self.slave.player.playlist[current_index:] ] 
        result_model, dfr = application.resource_manager.post(uri, songIDs=unicode(song_ids))
        dfr.addCallbacks(update_model, self._autoplay_fail)
        return dfr


    def _stop_autoplay_session(self):
        uri = 'grooveshark:///autoplay/stop'
        result_model, dfr = application.resource_manager.post(uri)
        dfr.addErrback(grooveshark_error_message)
        # We return defer.succeed and ignore 'dfr', because
        # we want the changes to the UI to be instantenous independently
        # of how much time it takes Grooveshark to process closing of 
        # the session.
        return defer.succeed(None)

    def activate(self):
        """
        On pressing, the autoplay control switches between enabled and disabled
        """
        if not self.disabled:
            self.set_state(not self.get_state())
        return defer.succeed(None)


class AutoPlayHappyControl(Control):
    """
    Control for sending approval of the current autoplay song to the 
    grooveshark server
    
    Its L{slave} passed to the constructor must be of type
    L{elisa.plugins.poblesec.player_video.PlayerController}.
    """
    def __init__(self, *args):
        super(AutoPlayHappyControl, self).__init__(*args)


    caption = _("Approve")
    glyphs = \
            {STATE_NORMAL: 'elisa.plugins.grooveshark.happy_normal',
             STATE_SELECTED: 'elisa.plugins.grooveshark.happy_selected',
             STATE_PRESSED: 'elisa.plugins.grooveshark.happy_active',
            }

    def activate(self):
        uri = 'grooveshark:///autoplay/smile'
        current_track = self.slave.player.get_current_model()
        if current_track.autoplay_id is not None:
            result_model, dfr = application.resource_manager.post(uri, autoplaySongID=current_track.autoplay_id)
            dfr.addErrback(grooveshark_error_message)
        else:
            dfr = defer.succeed(None)

        return dfr

class AutoPlaySadControl(Control):
    """
    Control for sending disapproval of the current autoplay song to 
    the grooveshark server

    Its L{slave} passed to the constructor must be of type
    L{elisa.plugins.poblesec.player_video.PlayerController}.
    """
    def __init__(self, *args):
        self.disabled = False
        super(AutoPlaySadControl, self).__init__(*args)

    
    caption = _("Disapprove")
    glyphs = \
            {STATE_NORMAL: 'elisa.plugins.grooveshark.sad_normal',
             STATE_SELECTED: 'elisa.plugins.grooveshark.sad_selected',
             STATE_PRESSED: 'elisa.plugins.grooveshark.sad_active',
            }

    def activate(self):
        def go_to_next_song(result):
            self.slave.player.play_next()

        def set_disabled(result):
            self.disabled = False
        
        if self.disabled:
            return
        
        uri = 'grooveshark:///autoplay/frown'
        current_track = self.slave.player.get_current_model()
        if current_track.autoplay_id is not None:
            self.disabled = True
            result_model, dfr = application.resource_manager.post(uri, autoplaySongID = current_track.autoplay_id )
            dfr.addCallbacks(go_to_next_song, grooveshark_error_message)
            dfr.addBoth(set_disabled)
        else:
            dfr = defer.succeed(None)

        return dfr

