# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Anna Wojdel <awojdel@fluendo.com>

from elisa.core.media_uri import MediaUri
from elisa.core.components.resource_provider import ResourceProvider

from elisa.core import common
from elisa.core.utils import defer

from elisa.plugins.http_client.http_client import ElisaHttpClient
from elisa.plugins.http_client.extern.client_http import ClientRequest

from elisa.plugins.base.models.media import PlayableModel, RawDataModel
from elisa.plugins.base.models.image import ImageModel


from elisa.plugins.grooveshark.models import DummyModel
from elisa.plugins.grooveshark.models import GroovesharkFeedModel
from elisa.plugins.grooveshark.models import GroovesharkArtistsListModel, \
                                            GroovesharkArtistModel 
from elisa.plugins.grooveshark.models import GroovesharkAlbumsListModel, \
                                            GroovesharkAlbumModel 
from elisa.plugins.grooveshark.models import GroovesharkTracksListModel, \
                                            GroovesharkTrackModel
from elisa.plugins.grooveshark.models import GroovesharkPlaylistsListModel, \
                                            GroovesharkPlaylistModel
from elisa.plugins.grooveshark.models import GroovesharkLoginModel

from elisa.plugins.grooveshark.constants import *

from twisted import internet
from twisted.internet import task
from twisted.web2 import responsecode
from twisted.web2.stream import BufferedStream
from twisted.web2.stream import MemoryStream
from twisted.web.client import getPage

try:
    from json.decoder import JSONDecoder
    from json.encoder import JSONEncoder
except ImportError:
    from simplejson.decoder import JSONDecoder
    from simplejson.encoder import JSONEncoder

import copy
import base64

class LoginError(Exception):
    pass

class ErrorInResponse(Exception):
    
    grooveshark_response = None

    def initialize(self, response=None):
        super(ErrorInResponse, self).initialize()
        self.grooveshark_response = response

        return self.frooveshark_response

class GroovesharkResourceProvider(ResourceProvider):
    """
    A resource provider to get the Grooveshark music over HTTP.
    """

    api_key_uri = 'http://%s/%s' % (ELISA_AUTH_SERVER, ELISA_AUTH_PATH)
    img_uri = 'http://%s/(.*)' % IMG_SERVER
    local_uri = '%s(.*)' % LOCAL_SERVER 

    supported_uri = img_uri + '|' + local_uri + '|' + api_key_uri

    config_doc = {'username': \
                  'the user name used for login to grooveshark webservice', 
                  'password': \
                  'password used to perform login to grooveshark webservice'}
    default_config = {'username': '', 'password': ''}

    def initialize(self):
        self.session_key = None
        
        self.decoder = JSONDecoder()
        self.encoder = JSONEncoder()

        #FIXME: 
        #Currently, all requests to Grooveshark are done by GetPage function 
        #from twisted. We should use ElisaHttpClient instead
        self._img_client = ElisaHttpClient(host=IMG_SERVER)
        self._session_client = ElisaHttpClient(host=ELISA_AUTH_SERVER)

        def login_failed(failure):
            self.warning('Login to Grooveshark failed: %s', failure)

        def reset_login_dfr(result_or_failure):
            self._login_dfr = None

        username = self.config['username'] or None
        password = self.config['password'] or None

        self._login_dfr = self.login(username=username, password=password)
        self._login_dfr.addErrback(login_failed)
        self._login_dfr.addBoth(reset_login_dfr)

        return super(GroovesharkResourceProvider, self).initialize()


    def clean(self):
        """
        Close all the open HTTP connections.
        """
        if self._login_dfr is not None:
            try:
                self._login_dfr.cancel()
            except defer.AlreadyCalledError:
                pass

        dfr = defer.DeferredList([self._img_client.close(),
                                  self._session_client.close(),
                                  self.send_post_request(SESSION_DESTROY_REQ)],
                                  consumeErrors=True)

        def parent_clean(result):
            return super(GroovesharkResourceProvider, self).clean()

        dfr.addCallback(parent_clean)
        return dfr

    def login(self, username=None, password=None):
        """
        Login to the Grooveshark webservice. 

        @param username: login of the user
        @type username:  string
        @param password: password of the user
        @type password:  string

        @rtype:          L{twisted.internet.defer.Deferred}
        """

        login_uri = 'grooveshark:///login'
        self.login_data = GroovesharkLoginModel()

        if username is not None and password is not None:
            self.login_data.username = username
            self.login_data.password = password
        else:
            return defer.succeed(False)

        def login_successful(result_model, model):
            model.user_id = result_model.user_id
            model.success = True
            #If login was successful save username and password in config
            self.config['username'] = model.username
            self.config['password'] = model.password

            return model.success

        self.debug('logging in to grooveshark: %s %s', self.login_data.username, self.login_data.password)
        result_model, dfr = self.post(login_uri, login_data=self.login_data)
        dfr.addCallback(login_successful, model=self.login_data)

        return dfr

    def logout(self):
        """
        Logout from the grooveshark webservice
        """
        logout_uri = 'grooveshark:///logout'

        def reset_login_data(result_model):
            self.login_data.reset()
            self.config['username'] = ''
            self.config['password'] = ''

            return defer.succeed(None)

        def logout_failed(failure):
            self.warning('Logout from grooveshark failed: %s', failure.value.message)
            return defer.fail(failure)

        self.debug('logging out from grooveshark')
        result_model, dfr = self.post(MediaUri(logout_uri))
        dfr.addCallback(reset_login_data)
        dfr.addErrback(logout_failed)

        return dfr

    def get(self, uri, context_model=None):
        """
        get() method in GroovesharkResourceProvider is used to get simple data
        requests (images for the coverart) from Grooveshark websites and 
        session_key from Moovida server.

        All communication with Grooveshark servers which requires RPC 
        is handled in post() method.
        """

        if str(uri)==self.api_key_uri:
            http_client = self._session_client
            result_model = dict()
        elif uri.endswith('.png') | uri.endswith('.jpg'):
            http_client = self._img_client
            result_model = RawDataModel()
        else:
            return ( None, defer.fail( Exception(NotImplementedError) ) )

        url = str(uri)

        def response_read(response, model):
            if isinstance(model, RawDataModel):
                model.data = response
                model.size = len(response)
            elif isinstance(model, dict):
                model = self.decoder.decode(response)

            return model


        def request_done(response, model):
            if response.code == responsecode.OK:
                read_dfr = BufferedStream( response.stream ).readExactly()
                read_dfr.addCallback( response_read, model )
                return read_dfr
            elif response.code == responsecode.NOT_FOUND:
                return defer.fail(IOError('Resource not found at %s' % url))
            else:
                return defer.fail(Exception('Received an %d HTTP response code' % response.code))

        self.debug('sending request to %s', url)
        request_dfr = http_client.request(url)
        request_dfr.addCallback(request_done, result_model)
        return (result_model, request_dfr )


    def start_session(self):
        """
        The function that starts the Grooveshark session. It contacts the elisa
        session-key server in order to obtain a key for communicating with 
        grooveshark.
        """
        def request_fail(response):
            self.warning('Received response error: %s' % response)
            return defer.fail(Exception('Received response error: %s' % response))

        def set_sessionID(json_data):
            try:
                root = json_data['result']
                self.session_key = root['sessionID']
            except:
                self.warning(json_data['fault']['message'])
                return defer.fail(Exception(ErrorInResponse))            

        uri = MediaUri(self.api_key_uri)
        result_model, request_dfr = self.get(uri)
        request_dfr.addCallback(set_sessionID)
        request_dfr.addErrback(request_fail)

        return request_dfr

    def send_post_request(self, post_data ):
        """
        Send a request to the grooveshark server, and prepare the resulting 
        data for parsing. The post_data is prepared in post method.
        """
        def get_sessionID():
            if self.session_key == None:
                self.debug('starting session')
                return self.start_session()
            else:
                return defer.succeed(self.session_key)

        def send_request(sessionID, post_data):
            def request_fail(response):
                self.warning('Received response error: %s' % response.value)
                return defer.fail(Exception('Received response error: %s' % response.value))

            post_data['header']['sessionID'] = self.session_key

            post_str = self.encoder.encode(post_data)
            request_uri = MediaUri({'scheme': 'http', 
                                    'host': API_SERVER, 
                                    'path': API_PATH})

            self.debug('sending request: %s', post_str)
             
            #FIXME: change it back to HttpElisaClient
            request_dfr = getPage(str(request_uri), method='POST', \
                                  postdata=post_str.encode('UTF-8') )

            request_dfr.addCallback(self.decoder.decode)
            request_dfr.addErrback(request_fail)

            return request_dfr

        dfr = get_sessionID()
        dfr.addCallback(send_request, post_data)
        return dfr

    def post(self, uri, **postdata):
        """
        post() method is used for communicating with Grooveshark servers 
        using RPC. It is used to get the result based on given local URI, 
        such as:
        grooveshark:///popular/songs
        or
        grooveshark:///search
        Some of the handled URIs expect postdata to be filled (e.g. with 
        search query), some ignore it. They are all grouped here, because they 
        require the same general framework for sending RPC requests via 
        http-POST.
        """

        #Map from the provided URI to result model type and post request template.
        try:
            result_model = URI_REQUEST_MAP[str(uri)][0]()
            post_req = copy.deepcopy(URI_REQUEST_MAP[str(uri)][1])
        except KeyError:
            self.warning('Unknown uri: %s' % uri)
            return ( None, defer.fail(Exception(KeyError)) )

        #Fill in all the post-request empty fields in the 'parameters' section.
        try:          
            if 'parameters' in post_req:
                parameters = post_req['parameters']

                # Special case for user ID
                if 'userID' in parameters:
                    #parameters['userID'] = self.login_data.get_UserID()
                    parameters['userID'] = self.login_data.user_id

                # Special case for login credentials 
                if 'username' in parameters:
                    parameters['username'] = postdata['login_data'].username
                    parameters['password'] = postdata['login_data'].password

                #All other parameters map one-to-one to the keys passed in
                #postdata argument 
                for key in parameters.keys():
                    if parameters[key] is None:
                        parameters[key] = postdata[key]
        except KeyError:
            return ( None, defer.fail(Exception(KeyError)) )

        id_dfr = self.send_post_request(post_req)
        id_dfr.addCallback(self.parse_data, result_model)

        return (result_model, id_dfr )


    def parse_data(self, json_data, model):
        """
        This function chooses appropriate parsing method for the data 
        received from grooveshark server.
        """

        def iterate_artists(root):
            for item in root['artists']:
                artist = self._parse_into_artist(item, model)
                model.artists.append(artist)
                yield item

        def iterate_albums(root):
            for item in root['albums']:
                album = self._parse_into_album(item, model)
                model.albums.append(album)
                yield item

        def iterate_tracks(root):
            for item in root['songs']:
                track = self._parse_into_track(item, model)
                model.tracks.append(track)
                yield item

        def iterate_playlists(root):
            for item in root['playlists']:
                playlist = self._parse_into_playlist(item, model)
                model.playlists.append(playlist)
                yield item

        def return_model(result, model):
            return model

        if isinstance(model, PlayableModel):
            model.uri = MediaUri(json_data['result']['url'])
            model.stream_key = json_data['result']['streamKey']
            model.stream_server_id = json_data['result']['streamServerID']
            return model
        elif isinstance(model, GroovesharkLoginModel):
            try:
                model.user_id = json_data['result']['userID']
            except:
                self.warning(json_data['fault']['message'])
                return defer.fail(Exception(LoginError))
            return model
        else:
            try:
                root = json_data['result']
            except:
                self.warning(json_data['fault']['message'])
                return defer.fail(Exception(ErrorInResponse(json_data['fault']['message'])))

            if isinstance(model, GroovesharkArtistsListModel):
                dfr = task.coiterate(iterate_artists(root))
            elif isinstance(model, GroovesharkAlbumsListModel):
                dfr = task.coiterate(iterate_albums(root))
            elif isinstance(model, GroovesharkTracksListModel):
                dfr = task.coiterate(iterate_tracks(root))
            elif isinstance(model, GroovesharkPlaylistsListModel):
                dfr = task.coiterate(iterate_playlists(root))
            elif isinstance(model, GroovesharkTrackModel):
                model = self._parse_into_track(root['song'], model)
                model.autoplay_id = root['song']['autoplaySongID']
                dfr = defer.succeed(None)
            elif isinstance(model, DummyModel):
                return defer.succeed(model)
            else:
                defer.fail( Exception(NotImplementedError) ) 

            dfr.addCallback(return_model, model)
            return dfr

    def _parse_into_artist(self, item, model):
        artist = GroovesharkArtistModel()

        artist.name = item['artistName']
        artist.artist_id = item['artistID']

        return artist


    def _parse_into_album(self, item, model):
        album = GroovesharkAlbumModel()

        album.name = item['albumName']
        album.album_id = item['albumID']
        album.artist = item['artistName']
        album.artist_id = item['artistID']
        try:
            image_path = item['image']['medium']
            album.cover = ImageModel()
            album.cover.references.append(MediaUri(image_path))
        except:
            pass

        return album

    def _parse_into_track(self, item, model):
        track = GroovesharkTrackModel()

        track.title = item['songName']
        track.artist = item['artistName']
        track.album = GroovesharkAlbumModel()
        track.album.name = item['albumName']
        track.album.artist = track.artist
        track.song_id = item['songID']
        image_path = item['image']['medium']
        thumbnail = ImageModel()
        thumbnail.references.append(MediaUri(image_path))
        track.thumbnails.append(thumbnail)

        return track

    def _parse_into_playlist(self, item, model):
        playlist = GroovesharkPlaylistModel()

        playlist.name = item['playlistName']
        playlist.playlist_id = item['playlistID']
        playlist.num_songs = item['numSongs']

        return playlist


    def authenticate_user(self):
        if not self._login_dfr:
            return self.login_data.success
        else:
            return None

