# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Anna Wojdel <awojdel@fluendo.com>

"""
Unit tests for grooveshark resource provider.
"""

from twisted.trial.unittest import TestCase
from elisa.core.utils.misc import run_functional_tests_check

from elisa.core.media_uri import MediaUri

from elisa.plugins.grooveshark.resource_provider import GroovesharkResourceProvider

from elisa.plugins.grooveshark.models import GroovesharkArtistsListModel, \
                                             GroovesharkAlbumsListModel, \
                                             GroovesharkTracksListModel, \
                                             GroovesharkPlaylistsListModel, \
                                             GroovesharkArtistModel, \
                                             GroovesharkAlbumModel, \
                                             GroovesharkTrackModel, \
                                             GroovesharkPlaylistModel, \
                                             GroovesharkLoginModel, \
                                             DummyModel

from elisa.plugins.base.models.media import PlayableModel, RawDataModel

from elisa.plugins.grooveshark.constants import IMG_SERVER


from twisted.internet import defer
from twisted.internet import task

class TestGroovesharkResourceProvider(TestCase):
    """
    This test case tests the AppleTrailersResourceProvider class.
    """

    def setUp(self):
        run_functional_tests_check()
        def created(provider):
            self.provider = provider
            return provider.initialize()

        create_dfr = GroovesharkResourceProvider.create({})
        create_dfr.addCallback(created)
        return create_dfr

    def tearDown(self):
        return self.provider.clean()

    def test_get_wrong_uri(self):
        """
        Test trying to get data from not existing URI.
        """

        uri = 'http://idontexist/%s/picture.png' % IMG_SERVER
        result_model, request_defer = self.provider.get(MediaUri(uri))
        self.failUnlessFailure(request_defer, Exception)

        uri = 'http://ania.txt'
        result_model, request_defer = self.provider.get(MediaUri(uri))
        self.failUnlessFailure(request_defer, Exception)
        return request_defer


    def test_post_wrong_uri(self):
        """
        Test trying to post to not existing URI.
        """
        uri = 'grooveshark://idontexist'
        result_model, request_defer = self.provider.post(MediaUri(uri))
        self.failUnlessFailure(request_defer, Exception)
        return request_defer


    def test_api_without_parameters(self):
        """
        Test API calls that don't need parameters (different than session_key)
        """

        #uri, model, attributes
        api_settings = [ ('grooveshark:///popular/artists', GroovesharkArtistsListModel, 'artists' ),
                         ('grooveshark:///popular/albums', GroovesharkAlbumsListModel, 'albums' ),
                         ('grooveshark:///popular/songs', GroovesharkTracksListModel, 'tracks' )
                       ]

        def test_api(uri):
            result_model, api_dfr = self.provider.post(MediaUri(uri))
            return result_model, api_dfr

        def check_response(result_model, model, attribute):
            self.failUnless(isinstance(result_model, model))
            self.failUnless(hasattr(result_model, attribute))
            return result_model

        def iterate_api_calls(api_settings):
            for uri, model, attribute in api_settings:
                result_model, api_dfr = test_api(uri)
                api_dfr.addCallback(check_response, model, attribute)
                yield api_dfr

        dfr = task.coiterate(iterate_api_calls(api_settings))
        return dfr


    def test_api_with_correct_parameters(self):
        """
        Test API calls that need additional parameters except session_key
        """

        #userID=1591 - Tom Waits
        #albumID=118291 - Mule Variations by Tom Waits
        #songID=15193 - Lowside Of The Road
        #playlistID=1840251 - one of our playlists
        correct_api_settings = [ ('grooveshark:///artist/songs', GroovesharkTracksListModel, 'tracks', {'artistID': '1591'} ),
                                ('grooveshark:///album/songs', GroovesharkTracksListModel, 'tracks', {'albumID': '118291'} ),
                                ('grooveshark:///playlist/songs', GroovesharkTracksListModel, 'tracks', {'playlistID': '1840251'} ),
                                ('grooveshark:///search/artists', GroovesharkArtistsListModel, 'artists', {'query': 'Tom Waits'} ),
                                ('grooveshark:///search/albums', GroovesharkAlbumsListModel, 'albums', {'query': 'Alice'} ),
                                ('grooveshark:///search/songs', GroovesharkTracksListModel, 'tracks', {'query': 'Blue Valentines'} ),
                                ('grooveshark:///song/favorite', DummyModel, None, {'songID': '15193'} ),
                                ('grooveshark:///song/unfavorite', DummyModel, None, {'songID': '15193'} ),
                                ('grooveshark:///playlist/addsong', DummyModel, None, {'playlistID': '1840251', 'songID': '15193', 'position': '1'} ),
                                ('grooveshark:///playlist/removesong', DummyModel, None, {'playlistID': '1840251', 'position': '1'} ),
                                ('grooveshark:///user/favorites', GroovesharkTracksListModel, 'tracks', {'userID': '214763'} ),
                                ('grooveshark:///user/playlists', GroovesharkPlaylistsListModel, 'playlists', {'userID': '214763'} ),
                                ('grooveshark:///song/stream', PlayableModel, 'uri' ,{'songID': '15193'} )
                               ]


        def test_api(uri, **param):
            result_model, api_dfr = self.provider.post(MediaUri(uri), **param)
            return result_model, api_dfr

        #checking only one attribute in returned model 
        def check_response(result_model, model, attribute):
            self.failUnless(isinstance(result_model, model))
            if attribute is not None:
                self.failUnless(hasattr(result_model, attribute))
            return result_model

        def iterate_api_calls(api_settings):
            for uri, model, attribute, param in api_settings:
                result_model, api_dfr = test_api(uri, **param)
                api_dfr.addCallback(check_response, model, attribute)
                yield api_dfr

        def log_out(result):
            result_model, dfr = self.provider.logout()
            return dfr

        dfr = self.provider.login('maxelisa', 'fluendo')
        dfr.addCallback(lambda login: task.coiterate(iterate_api_calls(correct_api_settings)) )
        dfr.addCallback(lambda api: log_out)

        return dfr


    def test_api_with_incorrect_parameters(self):
        """
        Test API calls that need additional parameters except session_key
        """

        api_settings = [ ('grooveshark:///artist/songs', {'artistID': '-1'} ),
                         ('grooveshark:///artist/songs', {} ),
                         ('grooveshark:///album/songs', {'albumID': '-1'} ),
                         ('grooveshark:///album/songs', {'albumID': None} ),
                         ('grooveshark:///playlist/songs', {'playlistID': '-1'} ),
                         ('grooveshark:///playlist/songs', {'playlistID': None} ),
                         ('grooveshark:///playlist/addsong', {'playlistID': '-1', 'songID': '15193'} ),
                         ('grooveshark:///playlist/addsong', {'playlistID': '1840251', 'songID': '-1'} ),
                         ('grooveshark:///playlist/addsong', {} ),
                         ('grooveshark:///playlist/removesong', {'playlistID': '-1', 'songID': '15193'} ),
                         ('grooveshark:///playlist/removeong', {'playlistID': '1840251', 'songID': '-1'} ),
                         ('grooveshark:///playlist/removesong', {} ),
                         ('grooveshark:///search/artists', {'query': None} ),
                         ('grooveshark:///search/albums', {'query': None} ),
                         ('grooveshark:///search/songs', {'query': None} ),
                         ('grooveshark:///song/favorite', {'songID': '-1'} ),
                         ('grooveshark:///song/favorite', {'songID': 'aaa'} ),
                         ('grooveshark:///song/unfavorite', {'songID': '-1'} ),
                         ('grooveshark:///song/unfavorite', {'songID': 'aaa'} )
                       ]

        def iterate_api_calls(api_settings):
            for uri, param in api_settings:
                result_model, api_dfr = self.provider.post(MediaUri(uri), **param)
                self.failUnlessFailure(api_dfr, Exception)
                yield api_dfr

        def log_out(result):
            result_model, dfr = self.provider.logout()
            return dfr

        dfr = self.provider.login('maxelisa', 'fluendo')
        dfr.addCallback(lambda login: task.coiterate(iterate_api_calls(api_settings)) )
        dfr.addCallback(lambda api: log_out)
        
        return dfr


    def test_get_image(self):

        def check_response(result_model):
            self.failUnless(isinstance(result_model, RawDataModel))

        uri = 'http://beta.grooveshark.com/webincludes/img/defaultart/album/mdefault.png'
        result_model, dfr = self.provider.get(MediaUri(uri))
        dfr.addCallback(check_response)

        return dfr


    def test_wrong_sessionkey(self):
        """
        Test trying to call grooveshark API with not valid session_key
        """
        #set wrong session_key
        self.provider.session_key='11111111111111111111111111111111'

        result_model, request_dfr = self.provider.post(MediaUri('grooveshark:///popular/songs'))
        self.failUnlessFailure(request_dfr, Exception)

        return request_dfr


    def test_login(self):
        """
        Test a session.login(username, password) API call 
        """
        login_settings = [ ('maxelisa', 'fluendo', True),
                           ('wrong_username', 'wrong_password', False), 
                           ('', '', False),
                           (None, None, False)]

        def check_response(response, expected_response):
            self.failUnless(response == expected_response)
            self.failUnless(self.provider.login_data.success == expected_response)
            if expected_response:
                self.failIf(self.provider.login_data.user_id is None)
            else:
                self.failUnless (self.provider.login_data.user_id is None)
            return response

        def iterate_login(login_settins):
            for username, password, expected_response in login_settings:
                login_dfr = self.provider.login(username, password)
                login_dfr.addCallback(check_response, expected_response)
                yield login_dfr

        dfr = task.coiterate(iterate_login(login_settings))

        return dfr


    def test_logout(self):
        def check_response(response):
            self.failIf(self.provider.login_data.success)
            self.failUnless(self.provider.login_data.user_id is None)

        def log_out(result):
            dfr = self.provider.logout()
            return dfr

        dfr = self.provider.login('maxelisa', 'fluendo')
        dfr.addCallback(log_out)
        dfr.addCallback(check_response)

        return dfr


    def test_logout_fail(self):
        dfr = self.provider.logout()
        self.failUnlessFailure(dfr, Exception)

        return dfr


    def test_wrong_grooveshark_method(self):
        post_data = { "method": "pop.getSongs", 
                         "parameters": { "limit": "", 
                                         "page": "" },
                         "header": { "sessionID": None } }                                         

        def check_response(response):
            self.failUnless('fault' in response)

        dfr = self.provider.send_post_request(post_data)
        dfr.addCallback(check_response)

        return dfr

